from typing import Dict, List


def parse_commands(
        tags_file,  # type: Dict
        commands    # type: List[str]
):
    # type: (...) -> str

    parsed = ""
    for command in commands:
        if command in tags_file:
            tags = tags_file[command]
            parsed += parse_tags(tags)

    return bytes(parsed)


def parse_tags(
    tags,       # type: Dict
    key=None,   # type: str
    cols=56     # type: int
):
    # type: (...) -> str

    parsed = ""
    for tag in tags:
        if key and "SeparatorDash" == key:
            tag = (tag * cols)
        parsed += b"{}".format(tag)

    return parsed



import copy
import cfgtools
from typing import List

from util import find_key_value

from ._BumpBar import BumpBar
from ._KdsView import KdsView
from ._Modifier import Modifier


def get_first_of(it, defval=None):
    return next(it, defval)


class Kds(object):
    def __init__(self, kds_config, kds_views):
        # type: (cfgtools.Group, List[KdsView]) -> None
        self.name = kds_config.name
        self.title = ''
        self.language = ''
        self.bump_bar = None
        self.modifiers = []
        self.views = []

        self.parse_kds(kds_config, kds_views)

    def parse_kds(self, kds_config, kds_views):
        self.language = find_key_value(kds_config, "Language")
        self.title = find_key_value(kds_config, "Title")

        for group in kds_config.groups:
            if group.name == 'BumpBar':
                self.bump_bar = BumpBar(group)
            elif group.name == 'ModifierPrefixes':
                self.modifiers = self.get_modifiers(group)
            elif group.name == 'Views':
                for view_group in group.groups:
                    view_name = get_first_of(x for x in view_group.keys if x.name == "ViewName").values[0]
                    view = copy.deepcopy(get_first_of(x for x in kds_views if x.name == view_name))
                    view.name = view_group.name
                    view.display_mode = get_first_of(x for x in view_group.keys if x.name == "Display").values[0]
                    view.title = get_first_of(x for x in view_group.keys if x.name == "Title").values[0]
                    self.views.append(view)

                self.views[0].selected = True


    @staticmethod
    def get_modifiers(group):
        modifiers = []
        for modifier_key in group.keys:
            modifiers.append(Modifier(modifier_key))

        return modifiers



import serial

from abc import ABCMeta
from logging import Logger
from typing import Dict

from nperipherals_core.model.custom_exceptions import DeviceNotConnected
from nperipherals_core.model.device import Device


class SerialDevice(Device):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            logger,                 # type: Logger
            device_configurations   # type: Dict[unicode, unicode]
    ):
        # type: (...) -> None

        self.logger = logger
        self.names = device_configurations["Name"]
        self.port = str(device_configurations["Port"])
        self.bud_rate = int(device_configurations["BaudRate"])
        self.data_bits = int(device_configurations["DataBits"])
        self.serial = None

    def get_name(
            self
    ):
        # type: (...) -> unicode

        if isinstance(self.names, list):
            return ','.join(self.names)

        return self.names

    def initialize(
            self
    ):
        # type: (...) -> None

        try:
            if self.serial is None:
                stop_bits = serial.STOPBITS_ONE
                self.serial = serial.Serial(
                        port=self.port,
                        baudrate=self.bud_rate,
                        bytesize=self.data_bits,
                        timeout=2,
                        stopbits=stop_bits
                )
        except Exception as e:
            self.logger.exception("Error getting serial connection: [{}]".format(repr(e)))
            raise DeviceNotConnected()

    def connect(
            self
    ):
        # type: (...) -> None

        try:
            if self.serial is not None and self.serial.is_open is False:
                self.serial.open()
        except Exception as e:
            self.logger.exception("Error open serial port: [{}]".format(repr(e)))
            raise DeviceNotConnected()

    def disconnect(
            self
    ):
        # type: (...) -> None

        try:
            self.serial.close()
        except Exception as _:
            self.logger.exception("Error closing socket connection")



import time
from logging import getLogger

from modelcommon.printer import SeparatorTag
from nperipherals_core.interactor.print_interactor import parse_tags
from nperipherals_core.model.custom_exceptions import FailedWritingToDevice
from nperipherals_core.model.device import WriterDevice, ReaderDevice
from typing import Union, Dict

logger = getLogger(__name__)


class PrintInteractor(object):
    def __init__(
            self,
            writer_device,  # type: Union[ReaderDevice, WriterDevice]
            tags_file       # type: Dict
    ):
        # type: (...) -> None

        self.writer_device = writer_device
        self.tags_file = tags_file

    def print_data(
            self,
            data    # type: str
    ):
        # type: (...) -> None

        device_name = self.writer_device.get_name()
        write_retries = self.writer_device.get_write_retries()
        write_retry_sleep_in_seconds = self.writer_device.get_write_retry_sleep_in_seconds()

        logger.info("Writing to device [{}]".format(device_name))

        error_messages = []
        for i in range(write_retries):
            connected = False
            try:
                self.writer_device.initialize()
                self.writer_device.connect()
                connected = True
                data = self._replace_tags_print(data)
                self.writer_device.write(data)
                return
            except Exception as ex:
                error_messages.append("Exception #[{}]: [{}]".format(i, repr(ex)))
                logger.warning("Error writing to device [{}]".format(device_name), exc_info=True)
                time.sleep(write_retry_sleep_in_seconds)
            finally:
                if connected:
                    self.writer_device.disconnect()

        raise FailedWritingToDevice("|".join(error_messages))

    def _replace_tags_print(
            self,
            data    # type: str
    ):
        # type: (...) -> str

        while True:

            if b"\x00\x02" not in data:
                break
            try:
                start_index = data.index(b"{}".format(SeparatorTag.start))
                end_index = data.index(b"{}".format(SeparatorTag.end), start_index)
                key = data[start_index + 2:end_index].strip()
                cols = self.tags_file["ColsPrinter"] if "ColsPrinter" in self.tags_file else 56
                if key in self.tags_file:
                    parsed = parse_tags(self.tags_file[key], key, cols)
                    data = self._replacer(data, parsed, start_index, end_index)
                    continue

                data = self._replacer(data, "", start_index, end_index)
            except Exception as _:
                logger.exception("Error replace tags to print. Data: {}".format(data))
                break

        return data

    @staticmethod
    def _replacer(
            data,               # type: str
            replace_string,     # type: str
            start,              # type: int
            end                 # type: int
    ):
        # type: (...) -> str

        if start not in range(len(data)):
            raise ValueError("index outside given string")

        if start < 0:
            return replace_string + data
        if start > len(data):
            return data + replace_string

        return data[:start] + replace_string + data[end+2:]



from abc import ABCMeta, abstractmethod
from xml.etree import cElementTree

from ._Order import Order  # noqa
from typing import List, Optional, Union  # noqa
from ._Table import Table  # noqa


class TableService(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def list_tables(self, pos_id):
        # type: (Union[int, str]) -> List[Table]
        raise NotImplementedError()

    @abstractmethod
    def start_service(self, pos_id, table_id, seats, tab_id=""):
        # type: (Union[int, str], str, int, Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def recall_service(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def store_service(self, pos_id, table_id):
        # type: (Union[int, str], str) -> bool
        raise NotImplementedError()

    @abstractmethod
    def create_order(self, pos_id, table_id):
        # type: (Union[int, str], str) -> None
        raise NotImplementedError()

    @abstractmethod
    def get_table_picture(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def get_selected_table(self, pos_id):
        # type: (Union[int, str]) -> Table
        raise NotImplementedError()

    @abstractmethod
    def get_table_info(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def do_total_table(self, pos_id, table_id, tip_rate=None):
        # type: (Union[int, str], str, Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def reopen_table(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def close_table(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def set_table_available(self, pos_id, table_id):
        # type: (Union[int, str], str) -> None
        raise NotImplementedError()

    @abstractmethod
    def get_current_table_id(self, pos_id):
        # type: (Union[int, str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def clear_service_tenders(self, pos_id, table_id):
        # type: (Union[int, str], str) -> str
        raise NotImplementedError()

    @abstractmethod
    def change_service_tip(self, pos_id, table_id, percentage=None, amount=None):
        # type: (Union[int, str], str, Optional[str], Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def register_service_tender(self, pos_id, table_id, tender_id, amount, tender_details=""):
        # type: (Union[int, str], str, float, List[str], Optional[str]) -> str
        raise NotImplementedError()

    @abstractmethod
    def get_table_orders(self, pos_id, table_id):
        # type: (Union[int, str], str) -> List[Order]
        raise NotImplementedError()

    @abstractmethod
    def set_current_order(self, pos_id, order_id):
        # type: (Union[int, str], str) -> None
        raise NotImplementedError()

    @abstractmethod
    def get_service_seats_count(self, table_info_xml):
        # type: (cElementTree.XML) -> int
        raise NotImplementedError()

    @abstractmethod
    def set_sale_line_seat(self, pos_id, seat_number, order_id, line_number, item_id, level, part_code):
        # type: (Union[int, str], int, str, str, str, str, str) -> None
        raise NotImplementedError()



import logging

from bustoken import (
    TK_MAINTENANCE_TERMINATE_REQUEST,
    TK_MAINTENANCE_RESTART_REQUEST,
    TK_MAINTENANCE_RECREATE_FILES_XML,
    TK_MAINTENANCE_RECREATE_FILE_NUMBER,
    TK_MAINTENANCE_CLEAN_AUDIT_LOG,
)
from events import (
    CleanDatabasesProcessor,
    MaintenanceProcessor
)
from messagehandler import EventHandler
from msgbus import MBEasyContext, TK_SYS_ACK, TK_SYS_NAK
from systools import sys_log_exception


logger = logging.getLogger("Maintenance")


class MaintenanceEventHandler(EventHandler):
    def __init__(
        self,
        mb_context,                 # type: MBEasyContext
        maintenance_processor,      # type: MaintenanceProcessor
        clean_databases_processor   # type: CleanDatabasesProcessor
    ):
        # type: (...) -> None

        super(MaintenanceEventHandler, self).__init__(mb_context)
        self.mb_context = mb_context
        self.maintenance_processor = maintenance_processor
        self.clean_databases_processor = clean_databases_processor

    def subscribe_events(self, subject):
        return

    def terminate_event(self):
        return

    @staticmethod
    def get_handled_tokens():
        return [
            TK_MAINTENANCE_TERMINATE_REQUEST,
            TK_MAINTENANCE_RESTART_REQUEST,
            TK_MAINTENANCE_RECREATE_FILES_XML,
            TK_MAINTENANCE_RECREATE_FILE_NUMBER
        ]

    def handle_event(self, subject, _, __, ___):
        logger.info("Event received: [{}]".format(subject))

        if not self.maintenance_processor.is_inside_maintenance_window():
            logger.info("Not inside maintenance window. Exiting...")
            return

        if subject == "CleanDatabaseBackups":
            self.clean_databases_processor.clean_database_backups()
        elif subject == "CleanDatabases":
            self.clean_databases_processor.clean_databases()
            self.clean_databases_processor.vacuum_databases()

    def handle_message(self, msg):
        try:
            if msg.token == TK_MAINTENANCE_TERMINATE_REQUEST:
                response = self.maintenance_processor.handle_terminate_request(msg)
            elif msg.token == TK_MAINTENANCE_RESTART_REQUEST:
                response = self.maintenance_processor.handle_restart_request(msg)
            elif msg.token == TK_MAINTENANCE_CLEAN_AUDIT_LOG:
                response = self.maintenance_processor.clean_audit_log()
            else:
                response = (False, "")

            msg.token = TK_SYS_ACK if response[0] else TK_SYS_NAK
            self.mb_context.MB_ReplyMessage(msg, data=response[1])
        except Exception as ex:
            sys_log_exception("Unexpected exception handling event: {}".format(msg.token))
            msg.token = TK_SYS_NAK
            self.mb_context.MB_ReplyMessage(msg, data=str(ex))



from abc import ABCMeta
from logging import Logger
from typing import Dict, List, Union

from nperipherals_core.model.device import Device


class FileDevice(Device):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            logger,                 # type: Logger
            device_configurations   # type: Dict[unicode, unicode]
    ):
        # type: (...) -> None

        self.logger = logger
        self.names = device_configurations["Name"]

    def get_name(
            self
    ):
        # type: (...) -> unicode

        if isinstance(self.names, list):
            return ','.join(self.names)

        return self.names

    def initialize(
            self
    ):
        # type: (...) -> None

        pass

    def connect(
            self
    ):
        # type: (...) -> None

        pass

    def disconnect(
            self
    ):
        # type: (...) -> None

        pass



from logging import getLogger

from externals.device import FileDevice
from nperipherals_core.interactor.print_interactor import parse_commands
from nperipherals_core.model.device import WriterDevice, CashDrawer
from nperipherals_core.model.device_status import CashDrawerStatus, DeviceStatus
from typing import Dict

logger = getLogger(__name__)


class PrinterFileDevice(WriterDevice, FileDevice, CashDrawer):
    REQUEST_STATUS_CODE = b'\x10\x04\x01'
    SOCKET_BUF_SIZE = 1

    def __init__(
            self,
            device_configurations,  # type: Dict[unicode, unicode]
            tags_file               # type: Dict
    ):
        # type: (...) -> None

        super(PrinterFileDevice, self).__init__(logger, device_configurations)

        self.tags_file = tags_file
        self.pre_commands = parse_commands(self.tags_file, device_configurations.get("PrePrintCmds", []))
        self.post_commands = parse_commands(self.tags_file, device_configurations.get("PostPrintCmds", []))
        self.write_retries = int(device_configurations.get("Retries", 1))
        self.write_retry_sleep_in_seconds = float(device_configurations.get("RetriesSleepSeconds", 1.0))
        self.file_path = device_configurations["Path"]
        self.names = device_configurations["Name"]
        self.drawer_status = CashDrawerStatus.Closed

    @staticmethod
    def get_status(

    ):
        return DeviceStatus.Ok

    @staticmethod
    def get_drawer_status(

    ):
        # type: () -> CashDrawerStatus

        return CashDrawerStatus.Closed

    def write(
            self,
            data    # type: unicode
    ):
        # type: (...) -> None

        f = open(self.file_path, "a")
        f.write("##########START-PRINT##########\n")
        f.write(self.pre_commands)
        f.write(data)
        f.write(self.post_commands)
        f.write("\n##########PRINT-END############\n\n")
        f.close()

    def open_drawer(
            self
    ):
        # type: (...) -> None

        self.drawer_status = CashDrawerStatus.Opened

    def get_write_retries(self):
        return 1

    def get_write_retry_sleep_in_seconds(self):
        return 1
